/* add pixel x mask -> dst */

#ifdef BUILD_MMX
static void
init_add_pixel_mask_span_funcs_mmx(void)
{
}
#endif

#ifdef BUILD_MMX
static void
init_add_pixel_mask_pt_funcs_mmx(void)
{
}
#endif

/*-----*/

/* add_rel pixel x mask -> dst */

#ifdef BUILD_MMX
static void
init_add_rel_pixel_mask_span_funcs_mmx(void)
{
}
#endif

#ifdef BUILD_MMX
static void
init_add_rel_pixel_mask_pt_funcs_mmx(void)
{
}
#endif
