DEFINE_TYPE(actionslider)
DEFINE_TYPE(bg)
DEFINE_TYPE(bubble)
DEFINE_TYPE(button)
DEFINE_TYPE(calendar)
DEFINE_TYPE(check)
DEFINE_TYPE(clock)
DEFINE_TYPE(combobox)
DEFINE_TYPE(entry)
DEFINE_TYPE(fileselector)
DEFINE_TYPE(fileselector_button)
DEFINE_TYPE(fileselector_entry)
DEFINE_TYPE(frame)
DEFINE_TYPE(gengrid)
DEFINE_TYPE(genlist)
DEFINE_TYPE(hoversel)
DEFINE_TYPE(icon)
DEFINE_TYPE(index)
DEFINE_TYPE(label)
DEFINE_TYPE(list)
DEFINE_TYPE(map)
DEFINE_TYPE(multibuttonentry)
DEFINE_TYPE(naviframe)
DEFINE_TYPE(notify)
DEFINE_TYPE(panes)
DEFINE_TYPE(photocam)
DEFINE_TYPE(progressbar)
DEFINE_TYPE(radio)
DEFINE_TYPE(scroller)
DEFINE_TYPE(segment_control)
DEFINE_TYPE(slider)
DEFINE_TYPE(slideshow)
DEFINE_TYPE(spinner)
DEFINE_TYPE(thumb)
DEFINE_TYPE(toolbar)
DEFINE_TYPE(video)
DEFINE_TYPE(web)
