
EAPI void
elm_slideshow_item_show(Elm_Slideshow_Item *obj)
{
   elm_obj_slideshow_item_show(obj);
}

EAPI Efl_Canvas_Object *
elm_slideshow_item_object_get(const Elm_Slideshow_Item *obj)
{
   return elm_obj_slideshow_item_object_get(obj);
}
