#ifndef _EET_CXX_SUITE_H
#define _EET_CXX_SUITE_H

#include <cassert>
#include <algorithm>

#include <check.h>
#include "../efl_check.h"
void eet_cxx_test_descriptors(TCase* tc);

#endif /* _EET_CXX_SUITE_H */
