
EAPI void
evas_vg_shape_fill_set(Efl_Canvas_Vg_Shape *obj, Efl_Canvas_Vg_Node *f)
{
   efl_canvas_vg_shape_fill_set(obj, f);
}

EAPI Efl_Canvas_Vg_Node *
evas_vg_shape_fill_get(const Efl_Canvas_Vg_Shape *obj)
{
   return efl_canvas_vg_shape_fill_get(obj);
}

EAPI void
evas_vg_shape_stroke_fill_set(Efl_Canvas_Vg_Shape *obj, Efl_Canvas_Vg_Node *f)
{
   efl_canvas_vg_shape_stroke_fill_set(obj, f);
}

EAPI Efl_Canvas_Vg_Node *
evas_vg_shape_stroke_fill_get(const Efl_Canvas_Vg_Shape *obj)
{
   return efl_canvas_vg_shape_stroke_fill_get(obj);
}

EAPI void
evas_vg_shape_stroke_marker_set(Efl_Canvas_Vg_Shape *obj, Efl_Canvas_Vg_Node *m)
{
   efl_canvas_vg_shape_stroke_marker_set(obj, m);
}

EAPI Efl_Canvas_Vg_Node *
evas_vg_shape_stroke_marker_get(const Efl_Canvas_Vg_Shape *obj)
{
   return efl_canvas_vg_shape_stroke_marker_get(obj);
}
