#ifndef _EVAS_CONVERT_GRY_4_H
#define _EVAS_CONVERT_GRY_4_H

void evas_common_convert_rgba_to_8bpp_pal_gray64               (DATA32 *src, DATA8 *dst, int src_jump, int dst_jump, int w, int h, int dith_x, int dith_y, DATA8 *pal);

#endif /* _EVAS_CONVERT_GRY_4_H */
