#define EVAS_IMAGE_FOLDER "/resources/images"
#define EVAS_MODEL_FOLDER "/resources/models"
#define EVAS_VG_FOLDER "/resources/vg"

#define EVAS_CONVEX_HULL_FOLDER "/resources/models/convex_hull"
#define EVAS_SAVED_FILES "/saved_files"

#define EVAS_PROJECT_MODEL_FOLDER "/assets/models"
#define EVAS_PROJECT_IMAGE_FOLDER "/assets/images"
