/* add pixel x color --> dst */

static void
init_add_pixel_color_span_funcs_c(void)
{
}

static void
init_add_pixel_color_pt_funcs_c(void)
{
}

/*-----*/

/* add_rel pixel x color --> dst */

static void
init_add_rel_pixel_color_span_funcs_c(void)
{
}

static void
init_add_rel_pixel_color_pt_funcs_c(void)
{
}
