
EAPI void
elm_panes_fixed_set(Efl_Ui_Panes *obj, Eina_Bool fixed)
{
   efl_ui_panes_fixed_set(obj, fixed);
}

EAPI Eina_Bool
elm_panes_fixed_get(const Efl_Ui_Panes *obj)
{
   return efl_ui_panes_fixed_get(obj);
}
