#include <stdio.h>

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "Eo.h"

#include "colourablesquare_stub.h"
#include "ns_colourable.eo.h"
#include "ns_colourablesquare.eo.h"
#include "ns_colourablesquare.eo.c"
