body, table, div, p, dl {
  font-family: "Open Sans", sans-serif;
  font-size: 14px;
  line-height: 1.5;
}

body {
  background-color: #303030 ;
  color: #888888;
  margin: 0;
}

h1 {
  font-size: 150%;
}

h2 {
  font-size: 120%;
}

h3 {
  font-size: 100%;
}

dt {
  font-weight: bold;
}

ul {
  padding-left: 30px;
}

div.multicol {
  -moz-column-gap: 1em;
  -webkit-column-gap: 1em;
  -moz-column-count: 3;
  -webkit-column-count: 3;
}

p.startli, p.startdd, p.starttd {
  margin-top: 2px;
}

p.endli {
  margin-bottom: 0px;
}

p.enddd {
  margin-bottom: 4px;
}

p.endtd {
  margin-bottom: 2px;
}

caption {
  font-weight: bold;
}

span.legend {
  font-size: 70%;
  text-align: center;
}

h3.version {
  font-size: 90%;
  text-align: center;
}

div.navtab {
  background-color: #606060;
  border: 1px solid #303030;
  text-align: center;
}

div.qindex {
  background-color: #eeeeee;
  text-align: center;
  margin-left: auto;
  margin-right: auto;
  height: 18px;
  overflow: hidden;
  vertical-align: middle;
}

div.navpath {
  max-width: 980px;
  background-color: #606060;
  text-align: center;
  margin-left: auto;
  margin-right: auto;
  box-shadow: 0px 3px 5px #000000;
  height: 18px;
  overflow: hidden;
  vertical-align: middle;
}

div.navtab {
  margin-right: 15px;
}

a {
  color: #44aaff;
  font-weight: normal;
  text-decoration: none;
  text-shadow: 0px 0px 10px rgba(51, 153, 255, 0.5);
}

.contents a:visited {
  color: #2288ee;
}

a:hover {
  text-decoration: underline;
}

a.qindex {
  font-weight: bold;
}

a.qindexHL {
  font-weight: bold;
  background-color: #9CAFD4;
  color: #ffffff;
  border: 1px double #869DCA;
}

.contents a.qindexHL:visited {
  color: #ffffff;
}

a.el {
  font-weight: bold;
}

a.elRef {
}

a.code, a.code:visited {
  color: #4665A2;
}

a.codeRef, a.codeRef:visited {
  color: #4665A2;
}

dl.el {
  margin-left: -1cm;
}

.fragment {
  font-family: monospace, mono, fixed;
  font-size: 9pt;
  white-space: pre;
  border: 0;

  color: #bbbbbb;
  background-color: #202020;
  padding: 4px 6px;
  margin: 4px 8px 4px 2px;
  overflow: auto;
}

.line {
  font-family: monospace, mono, fixed;
  font-size: 9pt;
}

/* To avoid packed effect on bullet lists */
li .line:first-child {
   margin-top: 10px;
}

div.ah {
  background-color: #000000;
  font-weight: bold;
  color: #ffffff;
  margin-bottom: 3px;
  margin-top: 3px;
  padding: 0.2em;
}

div.groupHeader {
  margin-left: 16px;
  margin-top: 12px;
  font-weight: bold;
}

div.groupText {
  margin-left: 16px;
  font-style: italic;
}

td.indexkey {
  background-color: #eeeeee;
  font-weight: bold;
  border: 1px solid #dddddd;
  margin: 2px 0px 2px 0;
  padding: 2px 10px;
  white-space: nowrap;
  vertical-align: top;
}

td.indexvalue {
  background-color: #eeeeee;
  border: 1px solid #dddddd;
  padding: 2px 10px;
  margin: 2px 0px;
}

tr.memlist {
  background-color: #EEF1F7;
}

p.formulaDsp {
  text-align: center;
}

img.formulaDsp {
}

img.formulaInl {
  vertical-align: middle;
}

div.center {
  text-align: center;
  margin-top: 0px;
  margin-bottom: 0px;
  padding: 0px;
}

div.center img {
  border: 0px;
}

address.footer {
  text-align: right;
  padding-right: 12px;
}

img.footer {
  border: 0px;
  vertical-align: middle;
}

span.keyword {
  color: #4488cc
}

span.keywordtype {
  color: #4488cc
}

span.keywordflow {
  color: #e08000
}

span.comment {
  color: #3355bb
}

span.preprocessor {
  color: #66ff55
}

span.stringliteral {
  color: #ff8877
}

span.charliteral {
  color: #aa9922
}

span.vhdldigit {
  color: #ff00ff
}

span.vhdlchar {
  color: #000000
}

span.vhdlkeyword {
  color: #700070
}

span.vhdllogic {
  color: #ff0000
}

/*
.search {
  color: #003399;
  font-weight: bold;
}

form.search {
  margin-bottom: 0px;
  margin-top: 0px;
}

input.search {
  font-size: 75%;
  color: #000080;
  font-weight: normal;
  background-color: #e8eef2;
}
*/

td.tiny {
  font-size: 75%;
}

.dirtab {
  padding: 4px;
  border-collapse: collapse;
  border: 1px solid #A3B4D7;
}

th.dirtab {
  background: #EBEFF6;
  font-weight: bold;
}

hr {
  height: 0px;
  border: none;
  border-top: 1px solid #c9c6e6;;
}

hr.footer {
  height: 1px;
}

table.memberdecls {
  border-spacing: 0px;
  padding: 0px;
}

.mdescLeft, .mdescRight,
.memItemLeft, .memItemRight,
.memTemplItemLeft, .memTemplItemRight, .memTemplParams {
  background-color: #202020;
  border: none;
  margin: 4px;
  padding: 1px 0 0 8px;
}

.mdescLeft, .mdescRight {
  padding: 0px 8px 4px 8px;
  color: #555555;
}

.memItemLeft, .memItemRight, .memTemplParams {
  border-top: 1px solid #404040;
}

.memItemLeft, .memTemplItemLeft {
  white-space: nowrap;
}

.memItemRight {
  width: 100%;
}

.memTemplParams {
  color: #4665A2;
  white-space: nowrap;
}

.memtemplate {
  font-size: 80%;
  color: #4665A2;
  font-weight: normal;
  margin-left: 9px;
}

.memnav {
  background-color: #EBEFF6;
  border: 1px solid #A3B4D7;
  text-align: center;
  margin: 2px;
  margin-right: 15px;
  padding: 2px;
}

.mempage {
  width: 100%;
}

.memitem {
  padding: 0;
  margin-bottom: 10px;
  margin-right: 5px;
}

.memname {
  white-space: nowrap;
  margin-left: 6px;
}

.memproto, dl.reflist dt {
  padding: 6px 0px 6px 0px;
  background-color: #404040;
  color: #ffffff;
}

.memdoc, dl.reflist dd {
  padding: 2px 5px;
  background-color: #202020;
  border-top: 1px solid #404040;
  box-shadow: 5px 5px 5px rgba(0, 0, 0, 0.15);
  -moz-box-shadow: rgba(0, 0, 0, 0.15) 5px 5px 5px;
  -webkit-box-shadow: 5px 5px 5px rgba(0, 0, 0, 0.15);
}

dl.reflist dt {
  padding: 5px;
}

dl.reflist dd {
  margin: 0px 0px 10px 0px;
  padding: 5px;
}

.paramkey {
  text-align: right;
}

.paramtype {
  white-space: nowrap;
}

.paramname {
  color: #bbbbbb;
  white-space: nowrap;
}
.paramname em {
  font-style: normal;
}

.params, .retval, .exception, .tparams {
  border-spacing: 6px 2px;
}

.params .paramname, .retval .paramname {
  font-weight: bold;
  vertical-align: top;
}

.params .paramtype {
  font-style: italic;
  vertical-align: top;
}

.params .paramdir {
  font-family: mono, monospace, fixed;
  vertical-align: top;
}

.ftvtree {
  font-family: sans-serif;
  margin: 0px;
}

.directory {
  font-size: 9pt;
  font-weight: bold;
  margin: 5px;
}

.directory h3 {
  margin: 0px;
  margin-top: 1em;
  font-size: 11pt;
}

/*
.directory h3.swap {
  height: 61px;
  background-repeat: no-repeat;
  background-image: url("yourimage.gif");
}
.directory h3.swap span {
  display: none;
}
*/

.directory > h3 {
  margin-top: 0;
}

.directory p {
  margin: 0px;
  white-space: nowrap;
}

.directory div {
  display: none;
  margin: 0px;
}

.directory img {
  vertical-align: -30%;
}

.directory tr.even {
  background-color: #f2f2f2;
}

.directory tr {
  background-color: #ffffff;
}

.directory td.entry {
  width: 50%;
}

.directory-alt {
  font-size: 100%;
  font-weight: bold;
}

.directory-alt h3 {
  margin: 0px;
  margin-top: 1em;
  font-size: 11pt;
}

.directory-alt > h3 {
  margin-top: 0;
}

.directory-alt p {
  margin: 0px;
  white-space: nowrap;
}

.directory-alt div {
  display: none;
  margin: 0px;
}

.directory-alt img {
  vertical-align: -30%;
}

div.dynheader {
  margin-top: 8px;
}

address {
  font-style: normal;
  color: #2A3D61;
}

table.doxtable {
  border-collapse:collapse;
}

table.doxtable td, table.doxtable th {
  border: 1px solid #2D4068;
  padding: 3px 7px 2px;
}

table.doxtable th {
  background-color: #374F7F;
  color: #FFFFFF;
  font-size: 110%;
  padding-bottom: 4px;
  padding-top: 5px;
  text-align:left;
}

table.fieldtable {
  width: 100%;
  margin-bottom: 10px;
  border: 1px solid #aaaaaa;
  border-spacing: 0px;
  -moz-box-shadow: rgba(0, 0, 0, 0.15) 2px 2px 2px;
  -webkit-box-shadow: 2px 2px 2px rgba(0, 0, 0, 0.15);
  box-shadow: 2px 2px 2px rgba(0, 0, 0, 0.15);
}

.fieldtable td, .fieldtable th {
  padding: 3px 7px 2px;
}

.fieldtable td.fieldtype, .fieldtable td.fieldname {
  white-space: nowrap;
  border-right: 1px solid #aaaaaa;
  border-bottom: 1px solid #aaaaaa;
  vertical-align: top;
}

.fieldtable td.fielddoc {
  border-bottom: 1px solid #aaaaaa;
  width: 100%;
}

.fieldtable tr:last-child td {
  border-bottom: none;
}

.fieldtable th {
  background-color: #dddddd;
  color: #202020;
  padding-bottom: 4px;
  padding-top: 5px;
  text-align:left;
  border-bottom: 1px solid #aaaaaa;
}

.tabsearch {
  top: 0px;
  left: 10px;
  height: 36px;
  background-image: url('tab_b.png');
  z-index: 101;
  overflow: hidden;
  font-size: 13px;
}

.navpath ul {
  font-size: 9pt;
  height: 18px;
  color: #ffffff;
  overflow: hidden;
  margin: 0px;
  padding: 0px;
  vertical-align: middle;
}

.navpath li {
  list-style-type: none;
  float: left;
  padding-left: 10px;
  padding-right: 10px;
  border-right: 4px solid #3399ff;
  vertical-align: middle;
}

.navpath li.navelem a {
  height: 18px;
  display: block;
  text-decoration: none;
  outline: none;
}

.navpath li.navelem a:hover {
  color: #ffffff;
  text-shadow: 0px 0px 5px #3399ff;
}

.navpath li.footer {
  list-style-type:none;
  float: right;
  padding-left:10px;
  padding-right:15px;
  background-image:none;
  background-repeat:no-repeat;
  background-position:right;
  color:#364D7C;
  font-size: 9pt;
}

div.summary {
  float: right;
  font-size: 9pt;
  padding-right: 10px;
  padding-top: 10px;
  width: 50%;
  text-align: right;
  color: #ffffff;
  text-shadow: 0px 0px 5px #3399ff;
}

div.summary a {
  white-space: nowrap;
  color: #ffffff;
  font-weight: bold;
}

div.ingroups {
  margin-left: 5px;
  font-size: 9pt;
  padding-left: 5px;
  width: 50%;
  text-align: left;
}

div.ingroups a {
  white-space: nowrap;
  color: #ffffff;
}

.title {
  font-size: 24px;
  font-weight: bold;
  margin: 20px 2px;
  color: #ffffff;
  font-family: "Open Sans", sans-serif;
  font-weight: 500;
}

div.header, div.footer {
  margin-left: auto;
  margin-right: auto;
  max-width: 980px;
}

div.footer {
  font-size: 10px;
  text-align: center;
  border-top: 1px solid #404040;
}

div.footer span.copyright {
  padding-right: 50px;
  padding-left: 50px;
}

div.contents {
  margin-left: auto;
  margin-right: auto;
  max-width: 940px;
}

#projectlogo {
  text-align: center;
  vertical-align: middle;
  border-collapse: separate;
}

#projectlogo img {
  border: 0px none;
}

#projectname {
  font: 300% mono, monospace, fixed;
  margin: 0px;
  padding: 2px 0px;
}

#projectbrief {
  font: 120% Arial, Helvetica, sans-serif;
  margin: 0px;
  padding: 0px;
}

#top {
  background-color:  #EBE9FE;
}

#titlearea {
  box-shadow: 0px 3px 10px #000000;
  margin-left: auto;
  margin-right: auto;
  padding: 0px 10px 0px 10px;
}

#projectnumber {
  font: 50% Tahoma, Arial, sans-serif;
  margin: 0px;
  padding: 0px;
}

dl {
  padding: 0 0 0 10px;
}

dl.note, dl.warning, dl.attention, dl.pre, dl.post, dl.invariant, dl.deprecated, dl.todo, dl.test, dl.bug, dl.since {
  border-left: 4px solid;
  padding: 0 0 0 6px;
}

dl.note {
  border-color: #D0C000;
}

dl.warning, dl.attention {
  border-color: #FF0000;
}

dl.pre, dl.post, dl.invariant {
  border-color: #00D000;
}

dl.deprecated, dl.since {
  border-color: #505050;
}

dl.since dt {
  height: 0;
}

dl.todo {
  border-color: #00C0E0;
}

dl.test {
  border-color: #3030E0;
}

dl.bug {
  border-color: #C08050;
}

.image {
  text-align: center;
}

.dotgraph {
  text-align: center;
}

.mscgraph {
  text-align: center;
}

.caption {
  font-weight: bold;
}

div.zoom {
  border: 1px solid #90A5CE;
}

dl.citelist {
  margin-bottom: 50px;
}

dl.citelist dt {
  color: #334975;
  float: left;
  font-weight: bold;
  margin-right: 10px;
  padding: 5px;
}

dl.citelist dd {
  margin: 2px 0;
  padding: 5px 0;
}

@media print {
  #top { display: none; }
  #side-nav { display: none; }
  #nav-path { display: none; }
  body { overflow:visible; }
  h1, h2, h3, h4, h5, h6 { page-break-after: avoid; }
  .summary { display: none; }
  .memitem { page-break-inside: avoid; }
  #doc-content {
    margin-left:0 !important;
    height:auto !important;
    width:auto !important;
    overflow: inherit;
    display: inline;
  }
  pre.fragment {
    overflow: visible;
    text-wrap: unrestricted;
    white-space: -moz-pre-wrap;
    white-space: -pre-wrap;
    white-space: -o-pre-wrap;
    white-space: pre-wrap;
    word-wrap: break-word;
  }
}

.tabs, .tabs2, .tabs3 {
  margin-left: auto;
  margin-right: auto;
  background-color: #606060;
  padding: 5px 0px 5px 0px;
  text-align: center;
  box-shadow: 0px 3px 5px #000000;
  max-width: 980px;
  z-index: 101;
  font-size: 9pt;
}

.tabs2 {
  font-size: 9pt;
}

.tabs2 li {
  margin-top: 9pt;
}

.tabs3 {
  font-size: 9pt;
}

.tabs3 li {
  margin-top: 9pt;
}

.tablist {
  margin: 0;
  padding: 0;
  display: table;
}

.tablist li {
  float: left;
  display: table-cell;
  line-height: 24px;
  list-style: none;
  margin-left: 8px;
  margin-right: 8px;
}

.tablist a {
  box-shadow: 0px 1px 2px #000000;
  display: block;
  padding: 0px 5px 0px 5px;
  font-weight: bold;
  color: #ffffff;
  text-shadow: 0px 0px 5px #3399ff;
  text-decoration: none;
  outline: none;
}

.tabs3 .tablist a {
  padding: 0px 5px;
}

.tablist a:hover {
  color: #ffffff;
  text-shadow: 0px 1px 0px #000000;
  text-decoration: none;
  background-color: #808080;
}

.tablist li.current a {
  color: #202020;
  background-color: #a0a0a0;
  text-shadow: 0px 1px 0px #cccccc;
}


/* EdcRef page specific styles */

table.edcref {
  border-collapse:collapse;
  border: 1px solid #AAAAAA;
  margin-bottom: 30px;
}

table.edcref td.section a  {
  font-size: 2.0em;
}

table.edcref td.subsection  {
  background-color: #EEEEEE;
  border: 1px solid #AAAAAA;
}

table.edcref td.subsection a {
  font-size: 1.5em;
}

table.edcref td.property {
  vertical-align: top;
  width: 1%;
}

table.edcref .fragment {
  width: 90%;
  border: 1px solid #AAAAAA;
  background-color: #F6F6F6;
  padding-top: 10px;
  padding-left: 3px;
  margin-right: 100px;
  margin-top: 10px;
  margin-bottom: 10px;
}

table.edcref td.description {
  padding-right: 95px;
}

table.edcref td.property, table.edcref td.parameters  {
  font-weight: bold;
  background-color: #DDDDDD;
  border-top: 1px solid #AAAAAA;
  padding: 2px;
}

table.edcref_prop {
  width: 90%;
  border-collapse:collapse;
  border: 1px solid #AAAAAA;
  box-shadow: 5px 5px 5px rgba(0, 0, 0, 0.15);
}
