#ifndef _ELM_POPUP_ITEM_EO_H_
#define _ELM_POPUP_ITEM_EO_H_

#ifndef _ELM_POPUP_ITEM_EO_CLASS_TYPE
#define _ELM_POPUP_ITEM_EO_CLASS_TYPE

typedef Eo Elm_Popup_Item;

#endif

#ifndef _ELM_POPUP_ITEM_EO_TYPES
#define _ELM_POPUP_ITEM_EO_TYPES


#endif
/** Elementary popup item class
 *
 * @ingroup Elm_Popup_Item
 */
#define ELM_POPUP_ITEM_CLASS elm_popup_item_class_get()

EWAPI const Efl_Class *elm_popup_item_class_get(void) EINA_CONST;

#endif
