/*
 * Copyright 2019 by its authors. See AUTHORS.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///
/// @file eo_init.hh
///

#ifndef EFL_CXX_EO_INIT_HH
#define EFL_CXX_EO_INIT_HH

#include "eo_ops.hh"

namespace efl { namespace eo {

/// @addtogroup Efl_Cxx_API
/// @{

/// @brief A simple RAII implementation to initialize and terminate
/// the <em>EO Subsystem</em>.
///
struct eo_init
{
   /// @brief Default constructor.
   ///
   /// Invokes @c ::eo_init().
   ///
   eo_init() { detail::init(); }

   /// @brief Class destructor.
   ///
   /// Invokes @c ::eo_shutdown().
   ///
   ~eo_init() { detail::shutdown(); }
};

/// @}

} }

#endif // EFL_CXX_EO_INIT_HH
