/* add color -> dst */

static void
init_add_color_span_funcs_c(void)
{
}

static void
init_add_color_pt_funcs_c(void)
{
}

/*-----*/

/* add_rel color -> dst */

static void
init_add_rel_color_span_funcs_c(void)
{
}

static void
init_add_rel_color_pt_funcs_c(void)
{
}
