#ifdef HAVE_CONFIG_H
# include "elementary_config.h"
#endif


#include "elm_priv.h"
#include "Eio_Eo.h"

#include "elm_interface_fileselector.h"

#include "elm_interface_fileselector_eo.h"
#include "elm_interface_fileselector_eo.c"
