#ifndef SIMPLE_H
#define SIMPLE_H

typedef struct
{
   int a;
} Simple_Public_Data;

EAPI void simple_a_set(Eo *obj, int a);
EAPI int simple_a_get(Eo *obj);

EAPI void simple_a_set1(Eo *obj, int a);
EAPI void simple_a_set2(Eo *obj, int a);
EAPI void simple_a_set3(Eo *obj, int a);
EAPI void simple_a_set4(Eo *obj, int a);
EAPI void simple_a_set5(Eo *obj, int a);
EAPI void simple_a_set6(Eo *obj, int a);
EAPI void simple_a_set7(Eo *obj, int a);
EAPI void simple_a_set8(Eo *obj, int a);
EAPI void simple_a_set9(Eo *obj, int a);
EAPI void simple_a_set10(Eo *obj, int a);
EAPI void simple_a_set11(Eo *obj, int a);
EAPI void simple_a_set12(Eo *obj, int a);
EAPI void simple_a_set13(Eo *obj, int a);
EAPI void simple_a_set14(Eo *obj, int a);
EAPI void simple_a_set15(Eo *obj, int a);
EAPI void simple_a_set16(Eo *obj, int a);
EAPI void simple_a_set17(Eo *obj, int a);
EAPI void simple_a_set18(Eo *obj, int a);
EAPI void simple_a_set19(Eo *obj, int a);
EAPI void simple_a_set20(Eo *obj, int a);
EAPI void simple_a_set21(Eo *obj, int a);
EAPI void simple_a_set22(Eo *obj, int a);
EAPI void simple_a_set23(Eo *obj, int a);
EAPI void simple_a_set24(Eo *obj, int a);
EAPI void simple_a_set25(Eo *obj, int a);
EAPI void simple_a_set26(Eo *obj, int a);
EAPI void simple_a_set27(Eo *obj, int a);
EAPI void simple_a_set28(Eo *obj, int a);
EAPI void simple_a_set29(Eo *obj, int a);
EAPI void simple_a_set30(Eo *obj, int a);
EAPI void simple_a_set31(Eo *obj, int a);
EAPI void simple_a_set32(Eo *obj, int a);
EAPI void simple_a_get1(Eo *obj, int a);
EAPI void simple_a_get2(Eo *obj, int a);
EAPI void simple_a_get3(Eo *obj, int a);
EAPI void simple_a_get4(Eo *obj, int a);
EAPI void simple_a_get5(Eo *obj, int a);
EAPI void simple_a_get6(Eo *obj, int a);
EAPI void simple_a_get7(Eo *obj, int a);
EAPI void simple_a_get8(Eo *obj, int a);
EAPI void simple_a_get9(Eo *obj, int a);
EAPI void simple_a_get10(Eo *obj, int a);
EAPI void simple_a_get11(Eo *obj, int a);
EAPI void simple_a_get12(Eo *obj, int a);
EAPI void simple_a_get13(Eo *obj, int a);
EAPI void simple_a_get14(Eo *obj, int a);
EAPI void simple_a_get15(Eo *obj, int a);
EAPI void simple_a_get16(Eo *obj, int a);
EAPI void simple_a_get17(Eo *obj, int a);
EAPI void simple_a_get18(Eo *obj, int a);
EAPI void simple_a_get19(Eo *obj, int a);
EAPI void simple_a_get20(Eo *obj, int a);
EAPI void simple_a_get21(Eo *obj, int a);
EAPI void simple_a_get22(Eo *obj, int a);
EAPI void simple_a_get23(Eo *obj, int a);
EAPI void simple_a_get24(Eo *obj, int a);
EAPI void simple_a_get25(Eo *obj, int a);
EAPI void simple_a_get26(Eo *obj, int a);
EAPI void simple_a_get27(Eo *obj, int a);
EAPI void simple_a_get28(Eo *obj, int a);
EAPI void simple_a_get29(Eo *obj, int a);
EAPI void simple_a_get30(Eo *obj, int a);
EAPI void simple_a_get31(Eo *obj, int a);
EAPI void simple_a_get32(Eo *obj, int a);

extern const Efl_Event_Description _EV_A_CHANGED;
#define EV_A_CHANGED (&(_EV_A_CHANGED))

#define SIMPLE_CLASS simple_class_get()
const Efl_Class *simple_class_get(void);

#endif
