
EAPI void
elm_frame_collapse_set(Efl_Ui_Frame *obj, Eina_Bool collapse)
{
   efl_ui_frame_collapse_set(obj, collapse);
}

EAPI Eina_Bool
elm_frame_collapse_get(const Efl_Ui_Frame *obj)
{
   return efl_ui_frame_collapse_get(obj);
}

EAPI void
elm_frame_autocollapse_set(Efl_Ui_Frame *obj, Eina_Bool autocollapse)
{
   efl_ui_frame_autocollapse_set(obj, autocollapse);
}

EAPI Eina_Bool
elm_frame_autocollapse_get(const Efl_Ui_Frame *obj)
{
   return efl_ui_frame_autocollapse_get(obj);
}

EAPI void
elm_frame_collapse_go(Efl_Ui_Frame *obj, Eina_Bool collapse)
{
   efl_ui_frame_collapse_go(obj, collapse);
}
