#ifndef _ELM_CODE_WIDGET_LEGACY_EO_H_
#define _ELM_CODE_WIDGET_LEGACY_EO_H_

#ifndef _ELM_CODE_WIDGET_LEGACY_EO_CLASS_TYPE
#define _ELM_CODE_WIDGET_LEGACY_EO_CLASS_TYPE

typedef Eo Elm_Code_Widget_Legacy;

#endif

#ifndef _ELM_CODE_WIDGET_LEGACY_EO_TYPES
#define _ELM_CODE_WIDGET_LEGACY_EO_TYPES


#endif
/** Elementary code widget
 *
 * @ingroup Elm_Code_Widget_Legacy
 */
#define ELM_CODE_WIDGET_LEGACY_CLASS elm_code_widget_legacy_class_get()

EWAPI const Efl_Class *elm_code_widget_legacy_class_get(void) EINA_CONST;

#endif
