#ifndef _ELM_GENGRID_PAN_EO_H_
#define _ELM_GENGRID_PAN_EO_H_

#ifndef _ELM_GENGRID_PAN_EO_CLASS_TYPE
#define _ELM_GENGRID_PAN_EO_CLASS_TYPE

typedef Eo Elm_Gengrid_Pan;

#endif

#ifndef _ELM_GENGRID_PAN_EO_TYPES
#define _ELM_GENGRID_PAN_EO_TYPES


#endif
/** Elementary gengrid pan class
 *
 * @ingroup Elm_Gengrid_Pan
 */
#define ELM_GENGRID_PAN_CLASS elm_gengrid_pan_class_get()

EWAPI const Efl_Class *elm_gengrid_pan_class_get(void) EINA_CONST;

#endif
