#ifndef _ELM_WEB_NONE_EO_H_
#define _ELM_WEB_NONE_EO_H_

#ifndef _ELM_WEB_NONE_EO_CLASS_TYPE
#define _ELM_WEB_NONE_EO_CLASS_TYPE

typedef Eo Elm_Web_None;

#endif

#ifndef _ELM_WEB_NONE_EO_TYPES
#define _ELM_WEB_NONE_EO_TYPES


#endif
/** Elementary web module class
 *
 * @ingroup Elm_Web_None
 */
#define ELM_WEB_NONE_CLASS elm_web_none_class_get()

EWAPI const Efl_Class *elm_web_none_class_get(void) EINA_CONST;

#endif
