/* add pixel --> dst */

static void
init_add_pixel_span_funcs_c(void)
{
}

static void
init_add_pixel_pt_funcs_c(void)
{
}

/*-----*/

/* add_rel pixel --> dst */

static void
init_add_rel_pixel_span_funcs_c(void)
{
}

static void
init_add_rel_pixel_pt_funcs_c(void)
{
}
