#include "evas_common_private.h"
#include "evas_convert_gry_1.h"

void evas_common_convert_rgba_to_1bpp_gry_1_dith(DATA32 *src EINA_UNUSED, DATA8 *dst EINA_UNUSED, int src_jump EINA_UNUSED, int dst_jump EINA_UNUSED, int w EINA_UNUSED, int h EINA_UNUSED, int dith_x EINA_UNUSED, int dith_y EINA_UNUSED, DATA8 *pal EINA_UNUSED)
{
}
