
EAPI void
elm_separator_horizontal_set(Elm_Separator *obj, Eina_Bool horizontal)
{
   elm_obj_separator_horizontal_set(obj, horizontal);
}

EAPI Eina_Bool
elm_separator_horizontal_get(const Elm_Separator *obj)
{
   return elm_obj_separator_horizontal_get(obj);
}
